from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import User
from django.core import mail
from django.core.cache import cache
from django.test.utils import override_settings
from django.utils import six
from django.utils.datastructures import MultiValueDict
from dns import resolver as dns_resolver

from djblets.mail.dmarc import (DmarcPolicy, get_dmarc_record,
                                is_email_allowed_by_dmarc)
from djblets.mail.message import EmailMessage
from djblets.mail.testing import DmarcDnsTestsMixin
from djblets.mail.utils import (build_email_address,
                                build_email_address_for_user,
                                build_email_address_via_service)
from djblets.testing.testcases import TestCase


_CONSOLE_EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'


class DmarcTests(DmarcDnsTestsMixin, TestCase):
    """Tests for DMARC support."""

    def test_get_dmarc_record_with_no_record(self):
        """Testing get_dmarc_record with no DMARC record in DNS"""
        self.assertIsNone(get_dmarc_record('example.com', use_cache=False))

    def test_get_dmarc_record_with_empty(self):
        """Testing get_dmarc_record with empty DMARC record in DNS"""
        self.dmarc_txt_records['_dmarc.example.com'] = ''
        self.assertIsNone(get_dmarc_record('example.com', use_cache=False))

    def test_get_dmarc_record_with_org_domain(self):
        """Testing get_dmarc_record with falling back to organizational
        domain
        """
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1; p=reject;'

        record = get_dmarc_record('mail.corp.example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.REJECT)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'reject',
            })

    def test_get_dmarc_record_with_version_only(self):
        """Testing get_dmarc_record with DMARC record containing a DMARC
        version only
        """
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.UNSET)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 100)
        self.assertEqual(record.fields, {'v': 'DMARC1'})

    def test_get_dmarc_record_with_p_none(self):
        """Testing get_dmarc_record with DMARC record with p=none"""
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1; p=none;'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.NONE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'none',
            })

    def test_get_dmarc_record_with_p_quarantine(self):
        """Testing get_dmarc_record with DMARC record with p=quarantine"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine; pct=20'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 20)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'pct': '20',
            })

    def test_get_dmarc_record_with_p_reject(self):
        """Testing get_dmarc_record with DMARC record with p=reject"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=reject; pct=100'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.REJECT)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'reject',
                'pct': '100',
            })

    def test_get_dmarc_record_with_sp_none(self):
        """Testing get_dmarc_record with DMARC record with sp=none"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine; sp=none;'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.NONE)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'sp': 'none',
            })

    def test_get_dmarc_record_with_sp_quarantine(self):
        """Testing get_dmarc_record with DMARC record with sp=quarantine"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine; sp=quarantine;'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'sp': 'quarantine',
            })

    def test_get_dmarc_record_with_sp_reject(self):
        """Testing get_dmarc_record with DMARC record with p=reject"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine; sp=reject;'

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.REJECT)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'sp': 'reject',
            })

    def test_get_dmarc_record_with_different_whitespace(self):
        """Testing get_dmarc_record with DMARC record with different amounts
        of whitespace in record
        """
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1;p=quarantine;  sp = reject; '

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.REJECT)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'sp': 'reject',
            })

    def test_get_dmarc_record_with_bad_format(self):
        """Testing get_dmarc_record with DMARC record with bad format issues"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine; ; =value; a=b=c; sp='

        record = get_dmarc_record('example.com', use_cache=False)
        self.assertIsNotNone(record)
        self.assertEqual(record.hostname, 'example.com')
        self.assertEqual(record.policy, DmarcPolicy.QUARANTINE)
        self.assertEqual(record.subdomain_policy, DmarcPolicy.UNSET)
        self.assertEqual(record.pct, 100)
        self.assertEqual(
            record.fields,
            {
                'v': 'DMARC1',
                'p': 'quarantine',
                'a': 'b=c',
            })

    def test_get_dmarc_record_with_cache(self):
        """Testing get_dmarc_record with use_cache=True"""
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1; p=reject;'

        record1 = get_dmarc_record('example.com', use_cache=True)
        self.assertIsNotNone(record1)
        self.assertEqual(record1.hostname, 'example.com')
        self.assertEqual(
            record1.fields,
            {
                'v': 'DMARC1',
                'p': 'reject',
            })

        record2 = get_dmarc_record('example.com', use_cache=True)
        self.assertEqual(record1, record2)

        self.assertEqual(len(dns_resolver.query.spy.calls), 1)

    def test_is_email_allowed_by_dmarc_with_domain_policy_none(self):
        """Testing is_email_allowed_by_dmarc with domain policy=none"""
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1; p=none;'
        self.assertTrue(is_email_allowed_by_dmarc('test@example.com'))

    def test_is_email_allowed_by_dmarc_with_domain_policy_quarantine(self):
        """Testing is_email_allowed_by_dmarc with domain policy=quarantine"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=quarantine;'
        self.assertFalse(is_email_allowed_by_dmarc('test@example.com'))

    def test_is_email_allowed_by_dmarc_with_domain_policy_reject(self):
        """Testing is_email_allowed_by_dmarc with domain policy=reject"""
        self.dmarc_txt_records['_dmarc.example.com'] = 'v=DMARC1; p=reject;'
        self.assertFalse(is_email_allowed_by_dmarc('test@example.com'))

    def test_is_email_allowed_by_dmarc_with_domain_policy_pct_0(self):
        """Testing is_email_allowed_by_dmarc with domain 0% match"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=reject; pct=0;'
        self.assertTrue(is_email_allowed_by_dmarc('test@example.com'))

    def test_is_email_allowed_by_dmarc_with_subdomain_policy_none(self):
        """Testing is_email_allowed_by_dmarc with subdomain and policy=none"""
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=reject; sp=none;'
        self.assertTrue(is_email_allowed_by_dmarc('test@mail.example.com'))

    def test_is_email_allowed_by_dmarc_with_subdomain_policy_quarantine(self):
        """Testing is_email_allowed_by_dmarc with subdomain and
        policy=quarantine
        """
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=reject; sp=quarantine;'
        self.assertFalse(is_email_allowed_by_dmarc('test@mail.example.com'))

    def test_is_email_allowed_by_dmarc_with_subdomain_policy_reject(self):
        """Testing is_email_allowed_by_dmarc with subdomain and policy=reject
        """
        self.dmarc_txt_records['_dmarc.example.com'] = \
            'v=DMARC1; p=reject; sp=reject;'
        self.assertFalse(is_email_allowed_by_dmarc('test@mail.example.com'))


class EmailMessageTests(DmarcDnsTestsMixin, TestCase):
    """Tests for djblets.mail.message.EmailMessage."""

    def setUp(self):
        super(EmailMessageTests, self).setUp()

        mail.outbox = []

    def test_init_with_text_body_as_unicode(self):
        """Testing EmailMessage.__init__ with text_body=<unicode>"""
        # auto_generated is True by default, so test with that case to ensure
        # it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body=b'This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertEqual(email.body, 'This is a test.')
        self.assertIsInstance(email.body, six.text_type)

    def test_init_with_text_body_as_bytes(self):
        """Testing EmailMessage.__init__ with text_body=<bytes>"""
        # auto_generated is True by default, so test with that case to ensure
        # it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body=b'This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertEqual(email.body, 'This is a test.')
        self.assertIsInstance(email.body, six.text_type)

    def test_init_with_html_body_as_unicode(self):
        """Testing EmailMessage.__init__ with html_body=<unicode>"""
        # auto_generated is True by default, so test with that case to ensure
        # it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             html_body='<p>This is a test.</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertEqual(email.alternatives,
                         [('<p>This is a test.</p>', 'text/html')])
        self.assertIsInstance(email.alternatives[0][0], six.text_type)

    def test_init_with_html_body_as_bytes(self):
        """Testing EmailMessage.__init__ with html_body=<bytes>"""
        # auto_generated is True by default, so test with that case to ensure
        # it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             html_body=b'<p>This is a test.</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertEqual(email.alternatives,
                         [('<p>This is a test.</p>', 'text/html')])
        self.assertIsInstance(email.alternatives[0][0], six.text_type)

    def test_message_with_from(self):
        """Testing EmailMessage.message with from_email="""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             enable_smart_spoofing=False)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'], 'doc@example.com')
        self.assertEqual(email._headers['Reply-To'], 'doc@example.com')
        self.assertEqual(email._headers['Sender'], settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email._headers['X-Sender'],
                         settings.DEFAULT_FROM_EMAIL)

        msg = email.message()
        self.assertEqual(msg['From'], 'doc@example.com')
        self.assertEqual(msg['Reply-To'], 'doc@example.com')

    def test_message_with_sender(self):
        """Testing EmailMessage.message with sender="""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             html_body='<p>This is a test.</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             sender='noreply@example.com',
                             enable_smart_spoofing=False)

        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'], 'doc@example.com')
        self.assertEqual(email._headers['Sender'], 'noreply@example.com')
        self.assertEqual(email._headers['X-Sender'], 'noreply@example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'doc@example.com')
        self.assertEqual(msg['Sender'], 'noreply@example.com')
        self.assertEqual(msg['X-Sender'], 'noreply@example.com')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_from_spoofing_always(self):
        """Testing EmailMessage.message with from_spoofing=always"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             from_spoofing=EmailMessage.FROM_SPOOFING_ALWAYS)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'], 'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_from_spoofing_smart_and_allowed(self):
        """Testing EmailMessage.message with from_spoofing=smart and From
        address allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=none;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             from_spoofing=EmailMessage.FROM_SPOOFING_SMART)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'], 'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_from_spoofing_smart_and_not_allowed(self):
        """Testing EmailMessage.message with from_spoofing=smart and From
        address not allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=quarantine;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             from_spoofing=EmailMessage.FROM_SPOOFING_SMART)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_from_spoofing_never(self):
        """Testing EmailMessage.message with from_spoofing=never"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             from_spoofing=EmailMessage.FROM_SPOOFING_NEVER)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service',
                       DJBLETS_EMAIL_FROM_SPOOFING='always')
    def test_message_with_from_spoofing_setting_always(self):
        """Testing EmailMessage.message with
        settings.DJBLETS_EMAIL_FROM_SPOOFING=always
        """
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'], 'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service',
                       DJBLETS_EMAIL_FROM_SPOOFING='smart')
    def test_message_with_from_spoofing_setting_smart_and_allowed(self):
        """Testing EmailMessage.message with
        settings.DJBLETS_EMAIL_FROM_SPOOFING=smart and From address allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=none;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'], 'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service',
                       DJBLETS_EMAIL_FROM_SPOOFING='smart')
    def test_message_with_from_spoofing_setting_smart_and_not_allowed(self):
        """Testing EmailMessage.message with
        settings.DJBLETS_EMAIL_FROM_SPOOFING=smart and From address not allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=quarantine;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service',
                       DJBLETS_EMAIL_FROM_SPOOFING='never')
    def test_message_with_from_spoofing_setting_never(self):
        """Testing EmailMessage.message with
        settings.DJBLETS_EMAIL_FROM_SPOOFING=never
        """
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_enable_smart_spoofing_true_and_allowed(self):
        """Testing EmailMessage.message with enable_smart_spoofing=True and
        From address allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=none;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             enable_smart_spoofing=True)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_enable_smart_spoofing_true_and_not_allowed(self):
        """Testing EmailMessage.message with enable_smart_spoofing=True and
        From address not allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=quarantine;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             enable_smart_spoofing=True)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_message_with_enable_smart_spoofing_false(self):
        """Testing EmailMessage.message with enable_smart_spoofing=False"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'],
                             enable_smart_spoofing=False)

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'], 'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service',
                       EMAIL_ENABLE_SMART_SPOOFING=True)
    def test_message_with_smart_spoofing_setting_true_and_not_allowed(self):
        """Testing EmailMessage.message with
        settings.EMAIL_ENABLE_SMART_SPOOFING=True and
        From address not allowed
        """
        self.dmarc_txt_records['_dmarc.corp.example.com'] = \
            'v=DMARC1; p=quarantine;'

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='Doc Dwarf <doc@corp.example.com>',
                             sender='noreply@mail.example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('From', email._headers)
        self.assertNotIn('Sender', email.extra_headers)
        self.assertNotIn('X-Sender', email.extra_headers)
        self.assertIn('From', email.extra_headers)
        self.assertIn('Sender', email._headers)
        self.assertIn('X-Sender', email._headers)
        self.assertIn('Reply-To', email._headers)
        self.assertEqual(email.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(email.extra_headers['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(email._headers['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')
        self.assertEqual(email._headers['Sender'],
                         'noreply@mail.example.com')
        self.assertEqual(email._headers['X-Sender'],
                         'noreply@mail.example.com')

        msg = email.message()
        self.assertEqual(msg['From'],
                         'Doc Dwarf via My Service <noreply@mail.example.com>')
        self.assertEqual(msg['Reply-To'],
                         'Doc Dwarf <doc@corp.example.com>')

    def test_message_with_in_reply_to(self):
        """Testing EmailMessage.message with in_reply_to="""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             in_reply_to='someone@example.com')

        self.assertIn('In-Reply-To', email._headers)
        self.assertIn('References', email._headers)
        self.assertEqual(email._headers['In-Reply-To'], 'someone@example.com')
        self.assertEqual(email._headers['References'], 'someone@example.com')

        msg = email.message()
        self.assertEqual(msg['In-Reply-To'], 'someone@example.com')
        self.assertEqual(msg['References'], 'someone@example.com')

    def test_message_with_auto_generated_true(self):
        """Testing EmailMessage.message with auto_generated=True"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             auto_generated=True)

        self.assertIn('Auto-Submitted', email._headers)
        self.assertEqual(email._headers['Auto-Submitted'], 'auto-generated')

        msg = email.message()
        self.assertEqual(msg['Auto-Submitted'], 'auto-generated')

    def test_message_with_auto_generated_false(self):
        """Testing EmailMessage.message with auto_generated=False"""
        # auto_generated is True by default, so test with that case to ensure
        # it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('Auto-Submitted', email._headers)

        msg = email.message()
        self.assertNotIn('Auto-Submitted', msg)

    def test_message_with_prevent_auto_responses_true(self):
        """Testing EmailMessage.message with prevent_auto_responses=True"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             prevent_auto_responses=True)

        self.assertIn('X-Auto-Response-Suppress', email._headers)
        self.assertEqual(email._headers['X-Auto-Response-Suppress'],
                         'DR, RN, OOF, AutoReply')

        msg = email.message()
        self.assertEqual(msg['X-Auto-Response-Suppress'],
                         'DR, RN, OOF, AutoReply')

    def test_message_with_prevent_auto_responses_false(self):
        """Testing EmailMessage.message with prevent_auto_responses=False"""
        # prevent_auto_responses is False by default, so test with that case
        # to ensure it doesn't unintentionally change.
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'])

        self.assertNotIn('X-Auto-Response-Suppress', email._headers)

        msg = email.message()
        self.assertNotIn('X-Auto-Response-Suppress', msg)

    def test_message_with_extra_headers_dict(self):
        """Testing EmailMessage.message with extra headers as a dict"""
        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             html_body='<p>This is a test.</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             headers={
                                 'X-Foo': 'Bar',
                             })

        message = email.message()

        self.assertIn('X-Foo', message)
        self.assertEqual(message['X-Foo'], 'Bar')

    def test_message_with_extra_headers_multivalue_dict(self):
        """Testing EmailMessage.message with extra headers as a MultiValueDict
        """
        header_values = ['Bar', 'Baz']

        email = EmailMessage(subject='Test email',
                             text_body='This is a test.',
                             html_body='<p>This is a test.</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             headers=MultiValueDict({
                                 'X-Foo': header_values,
                             }))

        message = email.message()

        self.assertIn('X-Foo', message)
        self.assertEqual(set(message.get_all('X-Foo')),
                         set(header_values))

    def test_send_with_unicode_subject(self):
        """Testing EmailMessage.send with a Unicode subject"""
        email = EmailMessage(subject='\U0001f604',
                             text_body='This is a test',
                             html_body='<p>This is a test</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             sender='noreply@example.com')

        with self.settings(EMAIL_BACKEND=_CONSOLE_EMAIL_BACKEND):
            email.send()

    def test_send_with_unicode_body(self):
        """Testing EmailMessage.send with a Unicode body"""
        email = EmailMessage(subject='Test email',
                             text_body='\U0001f604',
                             html_body='<p>\U0001f604</p>',
                             from_email='doc@example.com',
                             to=['sleepy@example.com'],
                             sender='noreply@example.com')

        # If the e-mail doesn't send correctly, it will raise a
        # UnicodeDecodeError.
        with self.settings(EMAIL_BACKEND=_CONSOLE_EMAIL_BACKEND):
            email.send()


class UtilsTests(TestCase):
    """Unit tests for djblets.mail.utils."""

    def test_build_email_address_with_full_name(self):
        """Testing build_email_address with full name"""
        self.assertEqual(build_email_address(email='test@example.com',
                                             full_name='Test User'),
                         'Test User <test@example.com>')

    def test_build_email_address_without_full_name(self):
        """Testing build_email_address without full name"""
        self.assertEqual(build_email_address(email='test@example.com'),
                         'test@example.com')

    def test_build_email_address_for_user(self):
        """Testing build_email_address_for_user"""
        user = User.objects.create(username='test',
                                   first_name='Test',
                                   last_name='User',
                                   email='test@example.com')

        self.assertEqual(build_email_address_for_user(user),
                         'Test User <test@example.com>')

    def test_build_email_address_via_service(self):
        """Testing test_build_email_address_via_service"""
        self.assertEqual(
            build_email_address_via_service(
                email='test@example.com',
                full_name='Test User',
                service_name='My Service',
                sender_email='noreply@example.com'),
            'Test User via My Service <noreply@example.com>')

    def test_build_email_address_via_service_without_full_name(self):
        """Testing test_build_email_address_via_service without full name"""
        self.assertEqual(
            build_email_address_via_service(
                email='test@example.com',
                service_name='My Service',
                sender_email='noreply@example.com'),
            'test via My Service <noreply@example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME='My Service')
    def test_build_email_address_via_service_with_service_name_setting(self):
        """Testing test_build_email_address_via_service with service name
        from settings.EMAIL_DEFAULT_SENDER_SERVICE_NAME
        """
        self.assertEqual(
            build_email_address_via_service(
                email='test@example.com',
                full_name='Test User',
                sender_email='noreply@example.com'),
            'Test User via My Service <noreply@example.com>')

    @override_settings(EMAIL_DEFAULT_SENDER_SERVICE_NAME=None)
    def test_build_email_address_via_service_with_computed_service_name(self):
        """Testing test_build_email_address_via_service with service name
        computed from sender e-mail
        """
        self.assertEqual(
            build_email_address_via_service(
                email='test@example.com',
                full_name='Test User',
                sender_email='noreply@example.com'),
            '"Test User via example.com" <noreply@example.com>')

    @override_settings(DEFAULT_FROM_EMAIL='noreply@example.com')
    def test_build_email_address_via_service_with_sender_email_setting(self):
        """Testing test_build_email_address_via_service with sender e-mail
        from settings.DEFAULT_FROM_EMAIL
        """
        self.assertEqual(
            build_email_address_via_service(
                full_name='Test User',
                email='test@example.com',
                service_name='My Service'),
            'Test User via My Service <noreply@example.com>')



"""Middleware for HTTP-related processing."""

from __future__ import unicode_literals

import django


_uses_relative_redirects = (django.VERSION >= (1, 9))


class AbsoluteURLRedirectMiddleware(object):
    """Middleware that turns all relative URL redirects to absolute.

    Django 1.9 changed URL redirects to be relative by default (so long as
    they're redirecting to the same host). For compatibility across Django
    versions (and to theoretically work around some broken HTTP implementations
    out in the wild), this middleware will turn those relative URLs back into
    absolute URLs.
    """

    def process_response(self, request, response):
        """Process the HTTP response.

        If a relative redirect was returned, this will make it absolute.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            response (django.http.HttpResponse):
                The HTTP response to check.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response.
        """
        if (_uses_relative_redirects and
            'Location' in response and
            request.get_host()):
            # Convert the relative URL into an absolute one.
            response['Location'] = \
                request.build_absolute_uri(response['Location'])

        return response



from __future__ import unicode_literals

import threading
import weakref
from contextlib import contextmanager

import django
from django.db.models import F, Q
from django.db.models.signals import (m2m_changed, post_delete, post_save,
                                      pre_delete)
from django.utils import six

from djblets.db.fields.counter_field import CounterField


class InstanceState(weakref.ref):
    """Tracks state for a RelationCounterField instance assocation.

    State instances are bound to the lifecycle of a model instance.
    They're a type of weak reference for model instances that contain
    additional state needed for the tracking and update process.

    These are used for looking up the proper instance and
    RelationCounterFields on the other end of a reverse relation, given
    a model, relation name, and IDs, through the
    :py:attr:`RelationCounterField._saved_instance_states` or
    :py:attr:`RelationCounterField._unsaved_instance_states` or
    dictionaries.

    Instance states can either represent saved instances or unsaved
    instances. Unsaved instance states represent instances that haven't yet
    been saved to the database (with a primary key of ``None``). While saved
    instance states exist per-instance/relation name, there's only one
    unsaved instance state per instance.

    Once an unsaved instance is saved, new instance states will be stored
    for each field associated (which many turn into multiple states, as
    there's one per relation name). The old unsaved instance state is then
    discarded.
    """

    def __init__(self, model_instance):
        """Set up the state.

        Args:
            model_instance (django.db.models.Model):
                The model instance that this state tracks.
        """
        super(InstanceState, self).__init__(model_instance)

        self.to_clear = set()
        self.field_names = set()
        self._model_cls = type(model_instance)

    @property
    def model_instance(self):
        """The model instance being tracked.

        This will be ``None`` if the instance has been destroyed.
        """
        return self()

    def track_field(self, field):
        """Track information on a field referencing this state.

        Args:
            field (django.db.models.Field):
                The field to track.
        """
        self.field_names.add(field.attname)

    def __repr__(self):
        """Return a string representation of the instance state.

        Returns:
            unicode:
            A string representation listing the instance information.
        """
        model_instance = self.model_instance

        if model_instance is not None:
            return '<InstanceState for %s.pk=%s>' % (
                model_instance.__class__.__name__,
                model_instance.pk)
        else:
            return '<InstanceState for %r (destroyed)>' % self.model_cls


class RelationTracker(object):
    """Tracks relations and updates state for all affected CounterFields.

    This class is responsible for all the hard work of updating
    RelationCounterFields refererring to a relation, based on updates
    to that relation. It's really the meat of RelationCounterField.

    Each RelationTracker is responsible for a given model/relation name
    pairing, across all instances of a model and across all
    RelationCounterFields following that relation name.

    The main reason the code lives here instead of in each
    RelationCounterField is to keep state better in sync and to ensure
    we're only ever dealing with one set of queries per relation name.
    We're also simplifying signal registration, helping to make things
    less error-prone.
    """

    def __init__(self, model_cls, rel_field_name):
        self._rel_field_name = rel_field_name

        if django.VERSION >= (1, 8):
            # Django >= 1.8
            self._rel_field = model_cls._meta.get_field(rel_field_name)
            rel_model = self._rel_field.model
            is_rel_direct = (not self._rel_field.auto_created or
                             self._rel_field.concrete)
            is_m2m = self._rel_field.many_to_many
        else:
            # Django < 1.8
            self._rel_field, rel_model, is_rel_direct, is_m2m = \
                model_cls._meta.get_field_by_name(rel_field_name)

        self._is_rel_reverse = not is_rel_direct

        if not is_m2m and is_rel_direct:
            # This combination doesn't make any sense. There's only ever
            # one item on this side, so no point in counting. Let's just
            # complain about it.
            raise ValueError(
                "RelationCounterField cannot work with the forward end of "
                "a ForeignKey ('%s')"
                % rel_field_name)

        dispatch_uid = '%s-%s.%s-related-save' % (
            id(self),
            self.__class__.__module__,
            self.__class__.__name__)

        if is_m2m:
            # This is going to be one end or the other of a ManyToManyField
            # relation.
            if is_rel_direct:
                # This is a ManyToManyField, and we can get the 'remote_field'
                # or 'rel' attribute (depending on Django version) through it
                # below.
                m2m_field = self._rel_field
            else:
                # This is a RelatedObject. We need to get the field through
                # this.
                m2m_field = self._rel_field.field

            if hasattr(m2m_field, 'remote_field'):
                # Django >= 1.7
                remote_field = m2m_field.remote_field
            else:
                remote_field = m2m_field.rel

            if is_rel_direct:
                self._related_name = remote_field.related_name
            else:
                self._related_name = m2m_field.attname

            # Listen for all M2M updates on the through table for this
            # ManyToManyField. Unfortunately, we can't look at a
            # particular instance, but we'll use state tracking to do the
            # necessary lookups and updates in the handler.
            m2m_changed.connect(
                self._on_m2m_changed,
                weak=False,
                sender=remote_field.through,
                dispatch_uid=dispatch_uid)
        else:
            # This is a ForeignKey or similar. It must be the reverse end.
            assert not is_rel_direct

            model = self._get_rel_field_related_model(self._rel_field)
            self._related_name = self._rel_field.field.attname

            # Listen for deletions and saves on that model type. In the
            # handler, we'll look up state for the other end of the
            # relation (the side owning this RelationCounterField), so that
            # we can update the counts.
            #
            # Unfortunately, we can't listen on the particular instance, so
            # we use the state tracking.
            post_delete.connect(
                self._on_related_delete,
                weak=False,
                sender=model,
                dispatch_uid=dispatch_uid)
            post_save.connect(
                self._on_related_save,
                weak=False,
                sender=model,
                dispatch_uid=dispatch_uid)

    def _increment_fields(self, states, by=1):
        """Increment all associated fields' counters on instance states.

        Args:
            states (list of InstanceState):
                The instance states containing the model instance fields to
                increment.

            by (int, optional):
                The value to increment by.
        """
        with self._update_sync_fields(states) as (model_instance, field_names):
            RelationCounterField.increment_many(
                model_instance,
                {
                    field_name: by
                    for field_name in field_names
                })

    def _decrement_fields(self, states, by=1):
        """Decrement all associated fields' counters on instance states.

        Args:
            states (list of InstanceState):
                The instance states containing the model instance fields to
                decrement.

            by (int, optional):
                The value to decrement by.
        """
        with self._update_sync_fields(states) as (model_instance, field_names):
            RelationCounterField.decrement_many(
                model_instance,
                {
                    field_name: by
                    for field_name in field_names
                })

    def _zero_fields(self, states):
        """Zero out all associated fields' counters on instance states.

        Args:
            states (list of InstanceState):
                The instance states containing the model instance fields to
                zero out.
        """
        with self._update_sync_fields(states) as (model_instance, field_names):
            RelationCounterField._set_values(
                model_instance,
                {
                    field_name: 0
                    for field_name in field_names
                })

    def _reload_fields(self, states):
        """Reload all associated fields' counters on instance states.

        Args:
            states (list of InstanceState):
                The instance states containing the model instance fields to
                reload.
        """
        with self._update_sync_fields(states) as (model_instance, field_names):
            RelationCounterField._reload_model_instance(model_instance,
                                                        field_names)

    @contextmanager
    def _update_sync_fields(self, states):
        """Update field values and synchronize them to other model instances.

        This calculates a main state from the list of instance states,
        gathering the model instance and field names and yielding them as
        context to the calling method. After that method makes the field
        changes needed, this will synchronize those values to all other model
        instances from the other states passed.

        Args:
            states (list of InstanceState):
                The list of states to update.

        Yields:
            tuple of (InstanceState, list of unicode):
            The main model instance to work on, and the list of field names
            to update.
        """
        main_state = states[0]
        model_instance = main_state.model_instance

        if model_instance is not None:
            yield model_instance, main_state.field_names

            if len(states) > 1:
                self._sync_fields_from_main_state(main_state, states[1:])

    def _on_m2m_changed(self, instance, action, reverse, model, pk_set,
                        **kwargs):
        """Handler for when a M2M relation has been updated.

        This will figure out the necessary operations that may need to be
        performed, given the update.

        For post_add/post_remove operations, it's pretty simple. We see
        if there are any instances (by way of stored state) for any of the
        affected IDs, and we re-initialize them.

        For clear operations, it's more tricky. We have to fetch all
        instances on the other side of the relation before any database
        changes are made, cache them in the InstanceState, and then update
        them all in post_clear.
        """
        if reverse != self._is_rel_reverse:
            # This doesn't match the direction we're paying attention to.
            # Ignore it.
            return

        is_post_clear = (action == 'post_clear')
        is_post_add = (action == 'post_add')
        is_post_remove = (action == 'post_remove')

        if is_post_clear or is_post_add or is_post_remove:
            states = RelationCounterField._get_saved_states(
                type(instance), instance.pk, self._rel_field_name)

            if states:
                # Perform the database modifications with only the first
                # instance state. The rest will get reloaded later.
                if pk_set and is_post_add:
                    self._increment_fields(states, by=len(pk_set))
                elif pk_set and is_post_remove:
                    self._decrement_fields(states, by=len(pk_set))
                elif is_post_clear:
                    self._zero_fields(states)

                if not pk_set and is_post_clear:
                    # See the note below for 'pre_clear' for an explanation
                    # of why we're doing this.
                    main_state = states[0]
                    pk_set = main_state.to_clear
                    main_state.to_clear = set()

            if pk_set:
                # If any of the models have their own
                # RelationCounterFields, make sure they've been updated to
                # handle this side of things.
                if is_post_add:
                    update_by = 1
                else:
                    update_by = -1

                # Update all RelationCounterFields on the other side of the
                # relation that are referencing this relation.
                self._update_counts(model, pk_set, '_related_name',
                                    update_by)

                for pk in pk_set:
                    states = RelationCounterField._get_saved_states(
                        model, pk, self._related_name)

                    if states:
                        self._reload_fields(states)
        elif action == 'pre_clear':
            # m2m_changed doesn't provide any information on affected IDs
            # for clear events (pre or post). We can, however, look up
            # these IDs ourselves, and if they match any existing
            # instances, we can re-initialize their counters in post_clear
            # above.
            #
            # We do this by fetching the IDs (without instantiating new
            # models) and storing it in the associated InstanceState. We'll
            # use those IDs above in the post_clear handler.
            states = RelationCounterField._get_saved_states(
                instance.__class__, instance.pk, self._rel_field_name)

            if states:
                main_state = states[0]
                mgr = getattr(instance, self._rel_field_name)
                main_state.to_clear.update(mgr.values_list('pk', flat=True))

    def _on_related_delete(self, instance, **kwargs):
        """Handler for when a ForeignKey relation is deleted.

        This will check if a model entry that has a ForeignKey relation
        to this field's parent model entry has been deleted from the
        database. If so, any associated counter fields on this end will be
        decremented.
        """
        states = self._get_reverse_foreign_key_states(instance)

        if states:
            self._decrement_fields(states)
        else:
            self._update_unloaded_fkey_rel_counts(instance, -1)

    def _on_related_save(self, instance=None, created=False, raw=False,
                         **kwargs):
        """Handler for when a ForeignKey relation is created.

        This will check if a model entry has been created that has a
        ForeignKey relation to this field's parent model entry. If so, any
        associated counter fields on this end will be decremented.
        """
        if raw or not created:
            return

        states = self._get_reverse_foreign_key_states(instance)

        if states:
            self._increment_fields(states)
        else:
            self._update_unloaded_fkey_rel_counts(instance, 1)

    def _update_unloaded_fkey_rel_counts(self, instance, by):
        """Updates unloaded model entry counters for a ForeignKey relation.

        This will get the ID of the model being referenced by the
        matching ForeignKey in the provided instance. If set, it will
        update all RelationCounterFields on that model that are tracking
        the ForeignKey.
        """
        rel_pk = getattr(instance, self._rel_field.field.attname)

        if rel_pk is not None:
            self._update_counts(
                self._get_rel_field_parent_model(self._rel_field),
                [rel_pk], '_rel_field_name', by)

    def _update_counts(self, model_cls, pks, rel_attname, update_by):
        """Updates counts on all model entries matching the given criteria.

        This will update counts on all RelationCounterFields on all entries
        of the given model in the database that are tracking the given
        relation.
        """
        values = dict(
            (field.attname, F(field.attname) + update_by)
            for field in model_cls._meta.local_fields
            if (isinstance(field, RelationCounterField) and
                (getattr(field._relation_tracker, rel_attname) ==
                    self._rel_field_name))
        )

        if values:
            if len(pks) == 1:
                q = Q(pk=list(pks)[0])
            else:
                q = Q(pk__in=pks)

            model_cls.objects.filter(q).update(**values)

    def _sync_fields_from_main_state(self, main_state, other_states):
        """Synchronize field values across instances.

        This will take a main instance containing up-to-date values and
        synchronize those values to all other instances.

        Args:
            main_state (InstanceState):
                The main state to take values from.

            other_states (list of InstanceState):
                The other states to synchronize values to.
        """
        main_instance = main_state.model_instance

        if main_instance is not None:
            for other_state in other_states:
                other_instance = other_state.model_instance

                if other_instance is not None:
                    for field_name in other_state.field_names:
                        setattr(other_instance, field_name,
                                getattr(main_instance, field_name))
        else:
            # The instance fell out of scope. We'll have to just reload
            # all the other instances. This should be rare.
            for other_state in other_states:
                self._reload_fields(other_state)

    def _get_reverse_foreign_key_states(self, instance):
        """Return InstanceStates for the other end of a ForeignKey.

        This is used when listening to changes on models that establish a
        ForeignKey to this counter field's parent model. Given the instance
        on that end, we can get the state for this end.

        Args:
            instance (django.db.model.Model):
                The instance on the other end of the relation.

        Returns:
            list of InstanceState:
            The list of :py:class:`InstanceState`s for each instance on this
            end of the relation.
        """
        return RelationCounterField._get_saved_states(
            self._get_rel_field_parent_model(self._rel_field),
            getattr(instance, self._rel_field.field.attname),
            self._rel_field_name)

    def _get_rel_field_parent_model(self, rel_field):
        """Return the model owning a relation field.

        This provides compatibility across different versions of Django.
        """
        if hasattr(rel_field, 'parent_model'):
            # Django < 1.7
            return rel_field.parent_model
        else:
            # Django >= 1.7
            return rel_field.model

    def _get_rel_field_related_model(self, rel_field):
        """Return the model on the other side of a relation field.

        This provides compatibility across different versions of Django.
        """
        if hasattr(rel_field, 'related_model'):
            # Django >= 1.7
            return rel_field.related_model
        else:
            # Django < 1.7
            return rel_field.model


class RelationCounterField(CounterField):
    """A field that provides an atomic count of a relation.

    RelationCounterField is a specialization of CounterField that tracks
    how many objects there are on the other side of a ManyToManyField or
    ForeignKey relation.

    RelationCounterField takes the name of a relation (either a field name,
    for a forward ManyToManyField relation, or the "related_name" for
    the reverse relation of another model's ForeignKey or ManyToManyField.
    (Note that using a forward ForeignKey relation is considered invalid,
    as the count can only be 1 or 0.)

    The counter will be initialized with the number of objects on the
    other side of the relation, and this will be kept updated so long as
    all updates to the table are made using standard create/save/delete
    operations on models.

    Note that updating a relation outside of a model's regular API (such as
    through raw SQL or something like an update() call) will cause the
    counters to get out of sync. They would then need to be reset using
    ``reinit_{field_name}``.
    """

    # Stores state across all saved instances of a RelationCounterField.
    #
    # Django doesn't make it easy to track updates to the other side of a
    # relation, meaning we have to do it ourselves. This dictionary is keyed
    # off a tuple of (model_class, instance_pk, field_name) and maps to a
    # weakref.WeakValueDictionary mapping object IDs for model instances to
    # InstanceState objects (which are tied to the lifecycle of a particular
    # model instancee). These objects are used to look up model instances and
    # their RelationCounterFields.
    _saved_instance_states = {}

    # Stores instance states we're tracking whose instances haven't yet been
    # saved.
    #
    # An unsaved instance may never be saved, but we still need to listen to
    # events on it. To do this, we set up an InstanceState, like above,
    # keeping track of the information and signal connections related to it.
    # Unlike the above, there's only one level to this dictionary. It maps
    # object IDs to InstanceState objects.
    _unsaved_instance_states = {}

    # Most of the hard work really lives in RelationTracker below. Here, we
    # store all registered instances of RelationTracker. There will be one
    # per model_cls/relation_name pair.
    _relation_trackers = {}

    # A lock for handling instance state dictionary modification, to prevent
    # threads from stomping over each other.
    #
    # We use a reentrant lock here so that state updates that end up freeing
    # an InstanceState and then need a new lock during state resets won't
    # end up blocking.
    _state_lock = threading.RLock()

    # Flag for determining if global signal handlers need to be set up.
    _signals_setup = False

    @classmethod
    def has_tracked_states(cls):
        """Return whether there are currently any states being tracked.

        This will begin by cleaning up any expired states whose instances
        have been destroyed, if there are any. Then it will check if there
        are any remaining states still being tracked and return a result.

        Returns:
            bool:
            ``True`` if there are any states still being tracked.
            ``False`` if not.
        """
        cls._cleanup_state()

        return (bool(cls._saved_instance_states) or
                bool(cls._unsaved_instance_states))

    @classmethod
    def _cleanup_state(cls, instance_cls=None, instance_pk=None,
                       instance_id=None):
        """Clean up state for one or more instances.

        This will clear away any state tied to a destroyed instance, an
        instance with a given reference ID, or an instance with a given class
        and database ID. It's used to ensure that any old, removed entries
        (say, from a previous unit test, or when transitioning from an unsaved
        instance to saved) are cleared away before storing new state.

        Args:
            instance_cls (type, optional):
                The model class of the instance being removed.

            instance_pk (int, optional):
                The database ID of the instance (if known and if saved).

            instance_id (int, optional):
                The reference ID of the instance.
        """
        with cls._state_lock:
            cls._cleanup_state_for_dict(cls._unsaved_instance_states,
                                        instance_cls=instance_cls,
                                        instance_pk=instance_pk,
                                        instance_id=instance_id)

            # The saved instances dictionary is a dictionary full of
            # dictionaries, so we'll want to iterate through and see if
            # there's any keys in the outer dictionary that we need to
            # remove by the end.
            to_remove = []

            for key, states in six.iteritems(cls._saved_instance_states):
                cls._cleanup_state_for_dict(states,
                                            instance_cls=instance_cls,
                                            instance_pk=instance_pk,
                                            instance_id=instance_id)

                if not states:
                    # The states dictionary is now empty. Schedule it for
                    # removal.
                    to_remove.append(key)

            for key in to_remove:
                cls._saved_instance_states.pop(key, None)

    @classmethod
    def _cleanup_state_for_dict(cls, states, instance_cls, instance_pk,
                                instance_id):
        """Clean up state in a states dictionary.

        This is a utility function used by :py:meth:`_cleanup_state` for
        clearing out any state entries matching the given instance information
        or for those states that are no longer active.

        Args:
            states (list of InstanceState):
                The list of instance states to clean up.

            instance_cls (type, optional):
                The model class of the instance being removed.

            instance_pk (int, optional):
                The database ID of the instance (if known and if saved).

            instance_id (int, optional):
                The reference ID of the instance.
        """
        to_remove = []

        for key, state in six.iteritems(states):
            model_instance = state.model_instance

            if (model_instance is None or
                id(model_instance) == instance_id or
                (type(model_instance) is instance_cls and
                 model_instance.pk == instance_pk)):
                to_remove.append(key)

        for key in to_remove:
            states.pop(key, None)

    @classmethod
    def _store_state(cls, instance, field):
        """Store state for a model instance and field.

        This constructs an :py:class:`InstanceState` instance for the given
        model instance and :py:class:`RelationCounterField`. It then associates
        it with the model instance and stores it.

        If the instance has not yet been saved, the constructed state will be
        tied to the instance and stored in :py:attr:`_unsaved_instance_states`.
        If the instance has been saved, the constructed state will be tied to
        a combination of the instance and field relation name and stoerd in
        :py:attr:`_saved_instance_states`.

        Saved instances will have a :samp:`_{fieldname}_state` attribute stored
        that points to the :py:class:`InstanceState`, keeping the state's
        reference alive as long as the instance is alive.
        """
        cls._cleanup_state()

        with cls._state_lock:
            if instance.pk is None:
                states = cls._unsaved_instance_states
            else:
                main_key = (type(instance), instance.pk, field._rel_field_name)
                states = cls._saved_instance_states.setdefault(main_key, {})

            key = id(instance)

            try:
                state = states[key]
            except KeyError:
                state = InstanceState(instance)
                states[key] = state

            state.track_field(field)

            # Mark that this instance tracks RelationCounterField states,
            # so our signal handlers have something they can easily look for.
            instance._tracks_relcounterfield_states = True

            if instance.pk is not None:
                # Attach the state to the field. This, along with being stored
                # in a WeakValueDictionary above, ensures that, so long as the
                # field remains in memory, the state will remain as well, and
                # that the state will be removed once the field instance
                # disappears (which will happen when the model instance
                # disappears).
                #
                # There should never be more than one state attached for a
                # field, since store_state() is only called once per model
                # instance per field, and field names are unique.
                setattr(instance, '_%s_instance_state' % field.attname, state)

    @classmethod
    def _get_saved_states(cls, model_cls, instance_pk, rel_field_name):
        """Return instance states for the given parameters.

        The returned dictionary will contain a mapping of object IDs for
        each instance to the :py:class:`InstanceState` for each saved instance
        matching the model class, primary key, and field name.

        Args:
            model_cls (type):
                The model class of the instances to look up.

            instance_pk (int):
                The database ID of the instances to look up.

            rel_field_name (unicode):
                The name of the field relationship associated with the
                instances.

        Returns:
            list of InstanceState:
            A list of all alive instance states for the given criteria. The
            first is considered the "main" state for an operation.

            If no suitable instances are found, this will return ``None``.
        """
        key = (model_cls, instance_pk, rel_field_name)

        with cls._state_lock:
            states = cls._saved_instance_states.get(key)

            if states is not None:
                return [
                    state
                    for state in six.itervalues(states)
                    if state.model_instance is not None
                ]

        return None

    @classmethod
    def _on_instance_first_save(cls, instance=None, created=False, **kwargs):
        """Handler for the first save on a newly created instance.

        This will reset information on this instance, removing this
        existing state, and will then add new instance states for each
        field relation.

        Args:
            instance (django.db.models.Model):
                The model instance being saved.

            created (bool):
                Whether the object was created. This must always be
                true for this handler.

            **kwargs (dict):
                Extra keyword arguments passed to the handler.

        Returns:
            bool:
            ``True`` if this instance was handled. ``False`` if it was ignored.
        """
        if (instance is None or
            not instance.pk or
            not created or
            not getattr(instance, '_tracks_relcounterfield_states', False)):
            # This isn't an instance we're tracking. Ignore it.
            return False

        instance_id = id(instance)

        try:
            state = cls._unsaved_instance_states[instance_id]
        except KeyError:
            # This isn't a tracked unsaved instance. We can skip it.
            return

        model_instance = state.model_instance

        if model_instance is None:
            # The references dropped. We're no longer working with this state.
            return False

        assert instance is model_instance

        with cls._state_lock:
            # Remove the old state information from the dictionary. We could
            # do this with a .pop(key, None) above, but we don't want to lock
            # unnecessarily.
            #
            # Even though we just fetched it, we can't assume that the data
            # is still there, since another thread could have done something
            # with it, so we can't do a plain del().
            cls._unsaved_instance_states.pop(instance_id, None)

        # Now we can register each RelationCounterField on here.
        for field in type(instance)._meta.local_fields:
            if isinstance(field, cls):
                cls._store_state(instance, field)

        return True

    @classmethod
    def _on_instance_pre_delete(cls, instance=None, **kwargs):
        """Handler for when an instance is about to be deleted.

        This will reset the state of the instance, unregistering it from
        lists, and removing any pending signal connections.

        Args:
            instance (django.db.models.Model):
                The instance being deleted.

            **kwargs (dict):
                Extra keyword arguments passed to the handler.

        Returns:
            bool:
            ``True`` if this instance was handled. ``False`` if it was ignored.
        """
        if (instance is None or
            not getattr(instance, '_tracks_relcounterfield_states', False)):
            # This isn't an instance we're tracking. Ignore it.
            return False

        instance_id = id(instance)
        assert instance_id not in cls._unsaved_instance_states

        cls._cleanup_state(instance_cls=type(instance),
                           instance_pk=instance.pk,
                           instance_id=instance_id)

        return True

    def __init__(self, rel_field_name=None, *args, **kwargs):
        def _initializer(model_instance):
            if model_instance.pk:
                return getattr(model_instance, rel_field_name).count()
            else:
                return 0

        kwargs['initializer'] = _initializer

        super(RelationCounterField, self).__init__(*args, **kwargs)

        self._rel_field_name = rel_field_name
        self._relation_tracker = None

    def _do_post_init(self, instance):
        """Handle initialization of an instance of the parent model.

        This will begin the process of storing state about the model
        instance and listening to signals coming from the model on the
        other end of the relation.

        Args:
            instance (django.db.models.Model):
                The model instance being initialized.
        """
        super(RelationCounterField, self)._do_post_init(instance)

        cls = type(self)

        # The first time we reach this, we want to let the class itself begin
        # listening for signals for instance saves and deletes.
        #
        # Note that, in theory, this could end up being called twice, but
        # that's ultimately okay, since we'll be using the same dispatch UID
        # on all signal connections, and the signals will ensure in a
        # thread-safe way that there's only one callback per dispatch UID.
        if not cls._signals_setup:
            dispatch_uid = '%s.%s' % (cls.__module__, cls.__name__)
            post_save.connect(cls._on_instance_first_save,
                              dispatch_uid=dispatch_uid)
            pre_delete.connect(cls._on_instance_pre_delete,
                               dispatch_uid=dispatch_uid)
            cls._signals_setup = True

        cls._store_state(instance, self)

        if not self._relation_tracker:
            instance_cls = type(instance)
            key = (instance_cls, self._rel_field_name)

            try:
                self._relation_tracker = cls._relation_trackers[key]
            except KeyError:
                self._relation_tracker = RelationTracker(instance_cls,
                                                         self._rel_field_name)
                cls._relation_trackers[key] = self._relation_tracker



#
# models.py -- Extension models.
#
# Copyright (c) 2010-2011  Beanbag, Inc.
# Copyright (c) 2008-2010  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django.db import models
from django.utils.encoding import python_2_unicode_compatible

from djblets.db.fields import JSONField
from djblets.extensions.errors import InvalidExtensionError


@python_2_unicode_compatible
class RegisteredExtension(models.Model):
    """Extension registration info.

    An extension that was both installed and enabled at least once. This
    may contain settings for the extension.

    This does not contain full information for the extension, such as the
    author or description. That is provided by the Extension object itself.
    """
    class_name = models.CharField(max_length=128, unique=True)
    name = models.CharField(max_length=32)
    enabled = models.BooleanField(default=False)
    installed = models.BooleanField(default=False)
    settings = JSONField()

    def __str__(self):
        return self.name

    def get_extension_class(self):
        """Retrieves the python object for the extensions class."""
        if not hasattr(self, '_extension_class'):
            cls = None

            try:
                # Import the function here to avoid a mutual
                # dependency.
                from djblets.extensions.manager import get_extension_managers

                for manager in get_extension_managers():
                    try:
                        cls = manager.get_installed_extension(self.class_name)
                        break
                    except InvalidExtensionError:
                        continue
            except:
                return None

            self._extension_class = cls

        return self._extension_class

    extension_class = property(get_extension_class)

    class Meta:
        # Djblets 0.9+ sets an app label of "djblets_extensions" on
        # Django 1.7+, which would affect the table name. We need to retain
        # the old name for backwards-compatibility.
        db_table = 'extensions_registeredextension'



"""Unit tests for djblets.auth.forms.RegistrationForm."""

from __future__ import unicode_literals

from django.contrib.auth.models import User

from djblets.auth.forms import RegistrationForm
from djblets.testing.testcases import TestCase


class RegistrationFormTests(TestCase):
    """Unit tests for djblets.auth.forms.RegistrationForm."""

    def test_clean_password2_with_match(self):
        """Testing RegistrationForm.clean_password2 with password2 field
        matching password1 field
        """
        form = RegistrationForm(data={
            'username': 'test',
            'email': 'test@example.com',
            'password1': 'password',
            'password2': 'password',
        })
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['password1'], 'password')
        self.assertEqual(form.cleaned_data['password2'], 'password')

    def test_clean_password2_without_match(self):
        """Testing RegistrationForm.clean_password2 with password2 field
        not matching password1 field
        """
        form = RegistrationForm(data={
            'username': 'test',
            'email': 'test@example.com',
            'password1': 'password',
            'password2': 'bad-password',
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'password2': ['Passwords must match'],
        })

    def test_save(self):
        """Testing RegistrationForm.save"""
        form = RegistrationForm(data={
            'username': 'test',
            'email': 'test@example.com',
            'password1': 'password',
            'password2': 'password',
            'first_name': 'Test',
            'last_name': 'User',
        })
        self.assertTrue(form.is_valid())

        user = form.save()

        self.assertIsNotNone(user)
        self.assertIsNotNone(user.pk)
        self.assertEqual(user.username, 'test')
        self.assertEqual(user.email, 'test@example.com')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertTrue(user.check_password('password'))

    def test_save_with_username_taken(self):
        """Testing RegistrationForm.save with username already taken"""
        User.objects.create(username='test')

        form = RegistrationForm(data={
            'username': 'test',
            'email': 'test@example.com',
            'password1': 'password',
            'password2': 'password',
            'first_name': 'Test',
            'last_name': 'User',
        })
        self.assertTrue(form.is_valid())

        user = form.save()

        self.assertIsNone(user)
        self.assertIn('username', form.errors)
        self.assertEqual(form.errors['username'],
                         ['Sorry, this username is taken.'])



"""An avatar service for providing uploaded images."""

from __future__ import unicode_literals

import os
from hashlib import md5
from uuid import uuid4

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.storage import DefaultStorage
from django.forms import forms, widgets
from django.utils.translation import ugettext_lazy as _

from djblets.avatars.forms import AvatarServiceConfigForm
from djblets.avatars.services.base import AvatarService


class FileUploadServiceForm(AvatarServiceConfigForm):
    """The FileUploadService configuration form."""

    avatar_service_id = 'file-upload'

    js_view_class = 'Djblets.Avatars.FileUploadSettingsFormView'
    template_name = 'avatars/services/file_upload_form.html'

    avatar_upload = forms.FileField(
        required=True,
        widget=widgets.FileInput(attrs={
            'accept': 'image/png, image/jpeg, image/gif',
        }))

    MAX_FILE_SIZE = 1 * 1024 * 1024
    is_multipart = True

    def clean_file(self):
        """Ensure the uploaded file is an image of an appropriate size.

        Returns:
            django.core.files.UploadedFile:
            The uploaded file, if it is valid.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the file is too large or the incorrect MIME type.
        """
        f = self.cleaned_data['avatar_upload']

        if f.size > self.MAX_FILE_SIZE:
            raise ValidationError(_('The file is too large.'))

        content_type = f.content_type.split('/')[0]

        if content_type != 'image':
            raise ValidationError(_('Only images are supported.'))

        return f

    def save(self):
        """Save the file and return the configuration.

        Returns:
            dict:
            The avatar service configuration.
        """
        storage = DefaultStorage()
        username = self.user.username

        uploaded_file = self.cleaned_data['avatar_upload']
        file_hash = md5()

        for chunk in uploaded_file.chunks():
            file_hash.update(chunk)

        # In the case where the filename does not have an extension,
        # splitext(filename) will return (filename, '').
        file_path = storage.get_valid_name(
            '%s%s'
            % (self.service.get_unique_filename(username),
               os.path.splitext(self.cleaned_data['avatar_upload'].name)[1])
        )

        file_path = os.path.join(username[0], username[:2], file_path)

        if self.service.file_path_prefix:
            file_path = os.path.join(self.service.file_path_prefix,
                                     file_path)

        file_path = storage.save(file_path, uploaded_file)

        return {
            'file_path': file_path,
            'file_hash': file_hash.hexdigest(),
        }


class FileUploadService(AvatarService):
    """An avatar service for uploaded images."""

    avatar_service_id = 'file-upload'
    name = _('File Upload')

    config_form_class = FileUploadServiceForm

    @property
    def file_path_prefix(self):
        """The storage location for uploaded avatars.

        This will be prepended to the path of all uploaded files. By default,
        it is controlled by the :setting:`UPLOADED_AVATARS_PATH` setting.
        """
        return getattr(settings, 'UPLOADED_AVATARS_PATH',
                       os.path.join('uploaded', 'avatars'))

    def get_unique_filename(self, filename):
        """Create a unique filename.

        The unique filename will be the original filename suffixed with a
        generated UUID.

        Args:
            filename (unicode):
                The filename, excluding the extension.

        Returns:
            unicode:
            The unique filename.
        """
        return '%s__%s' % (filename, uuid4())

    def get_avatar_urls_uncached(self, user, size):
        """Return the avatar URLs for the requested user.

        Args:
            user (django.contrib.auth.models.User):
                The user whose avatar URLs are to be fetched.

            size (int):
                The size (in pixels) the avatar is to be rendered at.

        Returns
            dict:
            A dictionary containing the URLs of the user's avatars at normal-
            and high-DPI.
        """
        storage = DefaultStorage()
        settings_manager = self._settings_manager_class(user)
        configuration = \
            settings_manager.configuration_for(self.avatar_service_id)

        if not configuration:
            return {}

        return {
            '1x': storage.url(configuration['file_path'])
        }

    def cleanup(self, user):
        """Clean up the uploaded file.

        This will delete the uploaded file from the storage.

        Args:
            user (django.contrib.auth.models.User):
                The user.
        """
        settings_manager = self._settings_manager_class(user)
        configuration = settings_manager.configuration_for(
            self.avatar_service_id)

        try:
            del configuration['file_hash']
        except KeyError:
            pass

        settings_manager.save()

        storage = DefaultStorage()
        storage.delete(configuration['file_path'])

    def get_etag_data(self, user):
        """Return the ETag data for the user's avatar.

        Args:
            user (django.contrib.auth.models.User):
                The user.

        Returns:
            list of unicode:
            The uniquely identifying information for the user's avatar.
        """
        settings_manager = self._settings_manager_class(user)
        configuration = \
            settings_manager.configuration_for(self.avatar_service_id)

        file_hash = configuration.get('file_hash')

        if not file_hash:
            storage = DefaultStorage()
            file_hash = md5()

            with storage.open(configuration['file_path'], 'rb') as f:
                file_hash.update(f.read())

            configuration['file_hash'] = file_hash.hexdigest()
            settings_manager.save()

        return [self.avatar_service_id, file_hash]



from __future__ import unicode_literals

from django.conf import settings
from django.forms import widgets
from django.utils.html import format_html


class RecaptchaWidget(widgets.Widget):
    """A widget for rendering the reCAPTCHA form field."""

    def render(self, *args, **kwargs):
        """Render the reCAPTCHA form field.

        Args:
            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            django.utils.safestring.SafeText:
            The rendered reCAPTCHA widget.
        """
        return format_html(
            '<div class="g-recaptcha" data-sitekey="{0}"></div>',
            settings.RECAPTCHA_PUBLIC_KEY)



#
# forms.py -- Form classes useful for extensions
#
# Copyright (c) 2010-2011  Beanbag, Inc.
# Copyright (c) 2008-2010  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

import warnings

from djblets.forms.forms import KeyValueForm
from djblets.util.decorators import cached_property


class SettingsForm(KeyValueForm):
    """Settings form for extension configuration.

    A base form for loading/saving settings for an extension. This is meant
    to be overridden by extensions to provide configuration pages. Any fields
    defined by the form will be loaded and saved automatically.
    """

    def __init__(self, extension, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        self.extension = extension
        self.settings = extension.settings

        super(SettingsForm, self).__init__(instance=extension.settings,
                                           *args, **kwargs)

    def set_key_value(self, key, value):
        """Set the value for an extension settings key.

        Args:
            key (unicode):
                The settings key.

            value (object):
                The settings value.
        """
        self.instance.set(key, value)

    def save_instance(self):
        """Save the instance."""
        self.instance.save()



"""Web API signals."""

from __future__ import unicode_literals

from django.core.signals import Signal


#: A signal indicating a WebAPI token has been created.
#:
#: Args:
#:     instance (djblets.webapi.models.WebAPIToken):
#:         The created instance.
#:
#:     auto_generated (bool):
#:         Whether or not the token was automatically generated.
webapi_token_created = Signal(providing_args=['instance', 'auto_generated'])


#: A signal indicating a WebAPI token has been updated.
#:
#: Args:
#:     instance (djblets.webapi.models.WebAPIToken):
#:         The updated instance.
webapi_token_updated = Signal(providing_args=['instance'])



from django.conf import settings

from djblets.util.decorators import simple_decorator


@simple_decorator
def add_root_url(url_func):
    """Decorates a function that returns a URL to add the SITE_ROOT."""
    def _add_root(*args, **kwargs):
        url = url_func(*args, **kwargs)

        if url[0] != '/':
            raise ValueError('Returned URL is not absolute')

        if hasattr(settings, 'SITE_ROOT'):
            return '%s%s' % (settings.SITE_ROOT, url[1:])
        else:
            return url

    return _add_root



"""Unit tests for djblets.extensions.hooks.BaseRegistryHook."""

from django.utils import six

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import ExtensionHookPoint, BaseRegistryHook
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.registries.registry import Registry
from djblets.testing.testcases import TestCase


class BaseRegistryHookTests(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.BaseRegistryHook."""

    class DummyRegistry(Registry):
        lookup_attrs = ('foo_id',)

    class DummyItem(object):
        def __init__(self, foo_id):
            self.foo_id = foo_id

    def setUp(self):
        super(BaseRegistryHookTests, self).setUp()

        self.registry = self.DummyRegistry()

        @six.add_metaclass(ExtensionHookPoint)
        class DummyRegistryHook(BaseRegistryHook):
            registry = self.registry

        class TestExtension(Extension):
            pass

        self.hook_cls = DummyRegistryHook

        self.extension = self.setup_extension(TestExtension)

    def test_hook_register(self):
        """Testing BaseRegistryHook item registration"""
        self.assertEqual(list(self.registry), [])

        item = self.DummyItem(123)
        self.hook_cls(self.extension, item)

        self.assertIn(item, self.registry)

    def test_hook_unregister(self):
        """Testing BaseRegistryHook item unregistration"""
        self.assertEqual(list(self.registry), [])
        item = self.DummyItem(123)
        self.hook_cls(self.extension, item)

        self.extension.shutdown()
        self.assertEqual(list(self.registry), [])



"""Unit tests for djblets.privacy.consent.hooks.ConsentRequirementHook."""

from __future__ import unicode_literals

from djblets.extensions.extension import Extension
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.privacy.consent import (BaseConsentRequirement,
                                     get_consent_requirements_registry)
from djblets.privacy.consent.hooks import ConsentRequirementHook
from djblets.testing.testcases import TestCase


class MyExtension(Extension):
    pass


class MyConsentRequirement(BaseConsentRequirement):
    requirement_id = 'my-requirement'
    name = 'My Requirement'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'


class ConsentRequirementHookTests(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.privacy.consent.hooks.ConsentRequirementHook."""

    extension_class = MyExtension

    def setUp(self):
        super(ConsentRequirementHookTests, self).setUp()

        self.registry = get_consent_requirements_registry()
        self.extension = self.setup_extension(MyExtension)
        self.consent_requirement = MyConsentRequirement()
        self.consent_requirement_id = self.consent_requirement.requirement_id

    def tearDown(self):
        super(ConsentRequirementHookTests, self).tearDown()

        # The requirement should definitely not be in the registry at this
        # point. Even if we left it there at the end of the test, the parent
        # tearDown() should have shut down the extension, removing the item.
        self.assertNotIn(self.consent_requirement, self.registry)

    def test_registration(self):
        """Testing ConsentRequirementHook registration"""
        self.assertIsNone(self.registry.get_consent_requirement(
            self.consent_requirement_id))

        ConsentRequirementHook(self.extension, self.consent_requirement)
        self.assertEqual(
            self.registry.get_consent_requirement(self.consent_requirement_id),
            self.consent_requirement)

    def test_unregistration(self):
        """Testing ConsentRequirementHook unregistration"""
        self.assertIsNone(self.registry.get_consent_requirement(
            self.consent_requirement_id))

        ConsentRequirementHook(self.extension, self.consent_requirement)
        self.assertEqual(
            self.registry.get_consent_requirement(self.consent_requirement_id),
            self.consent_requirement)

        self.extension.shutdown()
        self.assertIsNone(self.registry.get_consent_requirement(
            self.consent_requirement_id))



"""Mixins for providing reCAPTCHA validation support in forms.

See :ref:`using-recaptcha` for a guide on using reCAPTCHA validation.
"""

from __future__ import unicode_literals

import json
import logging

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.six.moves.urllib.error import URLError
from django.utils.six.moves.urllib.parse import urlencode
from django.utils.six.moves.urllib.request import urlopen
from django.utils.translation import ugettext as _

from djblets.recaptcha.widgets import RecaptchaWidget


class RecaptchaFormMixin(forms.Form):
    """A form mixin for providing reCAPTCHA verification.

    If other mixins are used, this should be the first in the list of base
    classes to ensure the reCAPTCHA field is the last.

    Users of this mixin must set the :py:attr:`request` attribute as the
    current :py:class:`~django.http.HttpRequest` instance before
    :py:meth:`clean` is called.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the mixin.

        Args:
            *args (tuple):
                Additional positional arguments to pass to the superclass
                constructor.

            **kwargs (dict):
                Additional keyword arguments to pass to the superclass
                constructor.
        """
        super(RecaptchaFormMixin, self).__init__(*args, **kwargs)
        self.fields['g-recaptcha-response'] = forms.CharField(
            required=False,
            widget=RecaptchaWidget)

    @property
    def verify_recaptcha(self):
        """Whether or not the reCAPTCHA is to be verified.

        Returns:
            bool
            Whether or not the reCAPTCHA is to be verified.
        """
        return True

    def clean(self):
        if self.verify_recaptcha:
            recaptcha_response = self.cleaned_data.get('g-recaptcha-response')

            if not recaptcha_response:
                raise ValidationError([
                    _("You must verify that you're not a robot by clicking "
                      "\"'I'm not a robot.\"")
                ])

            data = urlencode({
                'secret': settings.RECAPTCHA_PRIVATE_KEY,
                'response': recaptcha_response,
                'remote-ip': self.request.META.get('REMOTE_ADDR'),
            })

            try:
                resp = urlopen(
                    'https://www.google.com/recaptcha/api/siteverify',
                    data)

                payload = resp.read()
            except URLError as e:
                logging.exception('Could not make reCAPTCHA request: HTTP %s: '
                                  '%s',
                                  e.code, e.read())
                raise ValidationError([
                    _('Could not validate reCAPTCHA. Please contact an '
                      'administrator.'),
                ])

            try:
                payload = json.loads(payload)
            except ValueError:
                logging.exception('Could not parse JSON payload from %r',
                                  payload)
                raise ValidationError([
                    _('Could not validate reCAPTCHA. Please contact an '
                      'administrator.'),
                ])

            try:
                if not payload['success']:
                    raise ValidationError([
                        _('Invalid reCAPTCHA response.'),
                    ])
            except KeyError:
                logging.exception('No "success" key in reCAPTCHA payload %r',
                                  payload)
                raise ValidationError([
                    _('Could not validate reCAPTCHA. Please contact an '
                      'administrator.'),
                ])

        return super(RecaptchaFormMixin, self).clean()



"""Unit tests for the ExtensionResource mixin."""

from __future__ import unicode_literals

from django.test import RequestFactory

from djblets.extensions.extension import Extension
from djblets.extensions.manager import ExtensionManager
from djblets.extensions.resources import (ExtensionResource,
                                          ExtensionRootResourceMixin)
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase
from djblets.webapi.resources import RootResource, WebAPIResource


class TestChildResource1(WebAPIResource):
    """Test api resource item under MockResourceOne."""

    name = 'test_child_resource_item_1'
    uri_object_key = 'test_child_resource_item_1_id'


class TestChildResource2(WebAPIResource):
    """Test api resource item under MockResourceTwo."""

    name = 'test_child_resource_item_2'
    uri_object_key = 'test_child_resource_item_2_id'


class TestResource1(WebAPIResource):
    """Test api resource exposed by MockExtension."""

    name = 'test_resource_1'
    item_child_resources = [TestChildResource1()]


class TestResource2(WebAPIResource):
    """Test api resource exposed by MockExtension."""

    name = 'test_resource_2'
    item_child_resources = [TestChildResource2()]


class TestExtension(Extension):
    """Test extension for use by ExtensionTemplateTests."""

    resources = [TestResource1(), TestResource2()]


class TestRootResource(ExtensionRootResourceMixin, RootResource):
    """RootResource for testing ExtensionRootResourceMixin."""

    def get_extension_resource(self):
        """Return the mocked extension resource.

        Returns:
            djblets.extensions.resources.ExtensionResource:
            The mock extension resource.
        """
        return self.list_child_resources[0]


class ExtensionRootResourceMixinTests(ExtensionTestCaseMixin, TestCase):
    """Unit tests for the ExtensionRootResourceMixin."""

    extension_class = TestExtension

    def setUp(self):
        self.mgr = ExtensionManager('')
        self.ext_resource = ExtensionResource(self.mgr)
        self.root_resource = TestRootResource([self.ext_resource])
        self.request = RequestFactory().get('/')

        super(ExtensionRootResourceMixinTests, self).setUp()

    def test_generate_extension_uris_for_template(self):
        """Testing ExtensionRootResourceMixin generates URI templates when
        extensions are initialized
        """
        self.root_resource._generate_extension_uris_for_template(
            self.extension)
        actual_result = self.root_resource.get_uri_templates(
            self.request)
        self.assertEqual(actual_result, {
            'extensions': 'http://testserver/extensions/',
            'extension': 'http://testserver/extensions/{extension_name}/',
            'test_resource_1s': (
                'http://testserver/extensions/djblets.webapi.tests.'
                'test_extension_resource_mixin.'
                'TestExtension/test-resource-1s/'
            ),
            'test_resource_2s': (
                'http://testserver/extensions/djblets.webapi.tests.'
                'test_extension_resource_mixin.'
                'TestExtension/test-resource-2s/'
            ),
            'root': 'http://testserver/',
        })

    def test_remove_extension_uris_from_template(self):
        """Testing ExtensionRootResourceMixin removes URI templates when
        extensions are disabled
        """
        self.root_resource._generate_extension_uris_for_template(
            self.extension)
        self.root_resource._remove_extension_uris_from_template(
            self.extension)
        actual_result = self.root_resource.get_uri_templates(
            self.request)
        self.assertEqual(actual_result, {
            'extensions': u'http://testserver/extensions/',
            'root': 'http://testserver/',
            'extension': 'http://testserver/extensions/{extension_name}/',
        })



from __future__ import unicode_literals

import inspect
import logging
import warnings

from django import template
from django.utils import six
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe
from pipeline.conf import settings as pipeline_settings
from pipeline.templatetags.pipeline import JavascriptNode, StylesheetNode

from djblets.extensions.hooks import TemplateHook
from djblets.extensions.manager import get_extension_managers


logger = logging.getLogger(__name__)
register = template.Library()


class ExtensionStaticMediaNodeMixin(object):
    """Mixin for extension-specific static media rendering logic.

    This is used to change the behavior for how static media in extensions
    are rendered, allowing us to work with both development setups and
    installed packages.
    """

    def render_compressed(self, *args, **kwargs):
        """Render an extension media bundle to HTML.

        If Pipeline is disabled (typically on development setups), this will
        attempt to render the compressed files, as normal. However, if that
        fails (due to the source files not being available in a package) or
        Pipeline is enabled (typically on a production setup), the bundle's
        output file will be rendered instead.

        Args:
            *args (tuple):
                Positional arguments to pass to the rendering function.

            **kwargs (dict):
                Keyword arguments to pass to the rendering function.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for loading the static media.
        """
        rendered = ''

        if not pipeline_settings.PIPELINE_ENABLED:
            rendered = self.render_compressed_sources(*args, **kwargs)

        if not rendered:
            rendered = self.render_compressed_output(*args, **kwargs)

        return mark_safe(rendered)


class ExtensionJavascriptNode(ExtensionStaticMediaNodeMixin, JavascriptNode):
    """Template node for including extension-specific JavaScript.

    This will allow both extensions in development mode or installed via a
    package to be used on a page.
    """


class ExtensionStylesheetNode(ExtensionStaticMediaNodeMixin, StylesheetNode):
    """Template node for including extension-specific CSS.

    This will allow both extensions in development mode or installed via a
    package to be used on a page.
    """


@register.simple_tag(takes_context=True)
def template_hook_point(context, name):
    """Register a place where TemplateHooks can render to.

    Args:
        context (dict):
            The template rendering context.

        name (unicode):
            The name of the CSS bundle to render.

    Returns:
        django.utils.safetext.SafeText:
        The rendered HTML.
    """
    def _render_hooks():
        request = context['request']

        for hook in TemplateHook.by_name(name):
            try:
                if hook.applies_to(request):
                    context.push()

                    try:
                        yield (hook.render_to_string(request, context),)
                    except Exception as e:
                        logger.exception('Error rendering TemplateHook %r: %s',
                                         hook, e)

                    context.pop()

            except Exception as e:
                logger.exception('Error when calling applies_to for '
                                 'TemplateHook %r: %s',
                                 hook, e)

    return format_html_join('', '{0}', _render_hooks())


@register.simple_tag(takes_context=True)
def ext_static(context, extension, path):
    """Outputs the URL to the given static media file provided by an extension.

    This works like the {% static %} template tag, but takes an extension
    and generates a URL for the media file within the extension.

    This is meant to be used with
    :py:class:`djblets.extensions.staticfiles.ExtensionFinder`.
    """
    return extension.get_static_url(path)


def _render_bundle(context, node_cls, extension, name, bundle_type):
    try:
        return node_cls('"%s"' % extension.get_bundle_id(name)).render(context)
    except:
        logger.error('Unable to render %s bundle "%s" for extension "%s" '
                     '(%s). The extension may not be installed correctly. '
                     'Try disabling and re-installing the extension, and '
                     'checking the log for related errors.',
                     bundle_type, name, extension.info.name, extension.id)
        return ''


def _render_css_bundle(context, extension, name):
    """Render a given CSS bundle.

    Args:
        context (dict):
            The template rendering context.

        extension (djblets.extensions.extension.Extension):
            The extension instance.

        name (unicode):
            The name of the CSS bundle to render.

    Returns:
        django.utils.safetext.SafeText:
        The rendered HTML.
    """
    return _render_bundle(context, ExtensionStylesheetNode, extension,
                          name, 'CSS')


def _render_js_bundle(context, extension, name):
    """Render a given JavaScript bundle.

    Args:
        context (dict):
            The template rendering context.

        extension (djblets.extensions.extension.Extension):
            The extension instance.

        name (unicode):
            The name of the JS bundle to render.

    Returns:
        django.utils.safetext.SafeText:
        The rendered HTML.
    """
    return _render_bundle(context, ExtensionJavascriptNode, extension,
                          name, 'JS')


@register.simple_tag(takes_context=True)
def ext_css_bundle(context, extension, name):
    """Return HTML to import an extension's CSS bundle.

    Args:
        context (dict):
            The template rendering context.

        extension (djblets.extensions.extension.Extension):
            The extension instance.

        name (unicode):
            The name of the CSS bundle to render.

    Returns:
        django.utils.safetext.SafeText:
        The rendered HTML.
    """
    return _render_css_bundle(context, extension, name)


@register.simple_tag(takes_context=True)
def ext_js_bundle(context, extension, name):
    """Return HTML to import an extension's JavaScript bundle.

    Args:
        context (dict):
            The template rendering context.

        extension (djblets.extensions.extension.Extension):
            The extension instance.

        name (unicode):
            The name of the CSS bundle to render.

    Returns:
        django.utils.safetext.SafeText:
        The rendered HTML.
    """
    return _render_js_bundle(context, extension, name)


def _get_extension_bundles(extension_manager_key, context, bundle_attr,
                           default_bundles, renderer):
    """Yield media bundles that can be rendered on the current page.

    This will look through all enabled extensions and find any with static
    media bundles that should be included on the current page, as indicated
    by the context.

    All bundles marked "default" will be included, as will any with an
    ``apply_to`` field containing a URL name matching the current page.

    If a bundle has an ``include_bundles`` key, the referenced bundles will
    also be outputted. Note that this does not check for duplicates, and is
    not recursive.

    Args:
        extension_manager_key (unicode):
            The key for the extension manager for these bundles.

        context (django.template.Context):
            The template context.

        bundle_attr (unicode):
            The attribute name for the bundle on the extension class.

        default_bundles (unicode):
            A string containing a comma-separated list of bundles to always
            include.

        renderer (callable):
            The renderer function to call for each applicable bundle.

    Yields:
        tuple of (unicode or django.utils.safetext.SafeText):
        The HTML used to include the bundled content.
    """
    request = context['request']
    if not getattr(request, 'resolver_match', None):
        return

    requested_url_name = request.resolver_match.url_name
    default_bundles = set(default_bundles.split(','))

    for manager in get_extension_managers():
        if manager.key != extension_manager_key:
            continue

        for extension in manager.get_enabled_extensions():
            bundles = getattr(extension, bundle_attr, {})

            for bundle_name, bundle in six.iteritems(bundles):
                if (bundle_name in default_bundles or
                    requested_url_name in bundle.get('apply_to', [])):
                    for include_bundle in bundle.get('include_bundles', []):
                        yield renderer(context, extension, include_bundle)

                    yield (renderer(context, extension, bundle_name),)

        break


@register.simple_tag(takes_context=True)
def load_extensions_css(context, extension_manager_key,
                        default_bundles='default'):
    """Load all CSS bundles that can be rendered on the current page.

    This will include all "default" bundles and any with an ``apply_to``
    containing a URL name matching the current page.

    Args:
        context (django.template.Context):
            The template context.

        extension_manager_key (unicode):
            The key for the extension manager for these bundles.

        default_bundles (unicode):
            A string containing a comma-separated list of bundles to always
            include. Defaults to ``"default"``.

    Returns:
        django.utils.safestring.SafeText:
        The HTML used to include the bundled content.
    """
    return format_html_join('', '{0}\n', _get_extension_bundles(
        extension_manager_key=extension_manager_key,
        context=context,
        bundle_attr='css_bundles',
        default_bundles=default_bundles,
        renderer=_render_css_bundle))


@register.simple_tag(takes_context=True)
def load_extensions_js(context, extension_manager_key,
                       default_bundles='default'):
    """Load all JavaScript bundles that can be rendered on the current page.

    This will include all "default" bundles and any with an ``apply_to``
    containing a URL name matching the current page.

    Args:
        context (django.template.Context):
            The template context.

        extension_manager_key (unicode):
            The key for the extension manager for these bundles.

        default_bundles (unicode):
            A string containing a comma-separated list of bundles to always
            include. Defaults to ``"default"``.

    Returns:
        django.utils.safestring.SafeText:
        The HTML used to include the bundled content.
    """
    return format_html_join('', '{0}\n', _get_extension_bundles(
        extension_manager_key=extension_manager_key,
        context=context,
        bundle_attr='js_bundles',
        default_bundles=default_bundles,
        renderer=_render_js_bundle))


@register.inclusion_tag('extensions/init_js_extensions.html',
                        takes_context=True)
def init_js_extensions(context, extension_manager_key):
    """Initializes all JavaScript extensions.

    Each extension's required JavaScript files will be loaded in the page,
    and their JavaScript-side Extension subclasses will be instantiated.
    """
    request = context['request']
    url_name = request.resolver_match.url_name

    for manager in get_extension_managers():
        if manager.key == extension_manager_key:
            js_extensions = []

            for extension in manager.get_enabled_extensions():
                for js_extension_cls in extension.js_extensions:
                    js_extension = js_extension_cls(extension)

                    if js_extension.applies_to(url_name):
                        js_extensions.append(js_extension)

            js_extension_items = []

            for js_extension in js_extensions:
                arg_spec = inspect.getargspec(js_extension.get_model_data)

                if arg_spec.keywords is None:
                    warnings.warn(
                        '%s.get_model_data will need to take keyword '
                        'arguments. The old function signature is deprecated.'
                        % js_extension_cls.__name__)

                    model_data = js_extension.get_model_data()
                else:
                    model_data = js_extension.get_model_data(request=request)

                js_extension_items.append({
                    'js_extension': js_extension,
                    'model_data': model_data,
                })

            return {
                'url_name': url_name,
                'js_extension_items': js_extension_items,
            }

    return {}
